﻿// derived from code here http://www.informit.com/guides/content.aspx?g=dotnet&seqNum=626
//
using Microsoft.Win32;
using System;
using System.Collections;
using System.Collections.Generic;
using System.Linq;
using System.Text;

namespace RWTextEdit
{
    public class MRUList<TValue>: System.Collections.IEnumerable, IEnumerable<TValue>
    {
        public event OrderChangedEventHandler OrderChanged;
        public delegate void OrderChangedEventHandler(object sender, EventArgs e);

        protected virtual void OnOrderChanged()
        {
            if (OrderChanged != null)
                OrderChanged(this, new EventArgs());
        }

        public class MRUItem
        {
            private TValue _value;

            public MRUItem(TValue v)
            {
                _value = v;
            }

            public TValue Value
            {
                get { return _value; }
            }
        }

        public delegate bool MRUEqualityComparer(TValue v, TValue w);

        MRUEqualityComparer isEqualFunc;
        LinkedList<MRUItem> items;
        private int version;

        MRUList(int cap, MRUEqualityComparer compFunc)
        {
            items = new LinkedList<MRUItem>();
            isEqualFunc = compFunc;
            version = 0;
            MaxSize = cap;
        }

        public int MaxSize { get; set; }
        public int Count { get { return items.Count; } }

        public void Add(TValue value)
        {
            // Check to see if the key is already in the dictionary
            if (FindNode(value) != null)
            {
                throw new ArgumentException("An item with the same key already exists.");
            }

            // If the list is at capacity, remove the LRU item.
            if (items.Count == MaxSize)
            {
                items.RemoveLast();
            }

            // Create a node for this key/value pair
            LinkedListNode<MRUItem> newNode = new LinkedListNode<MRUItem>(new MRUItem(value));
            // Add to the items list
            items.AddFirst(newNode);
            ++version;
            OnOrderChanged();
        }

        public MRUItem Find(TValue value)
        {
            LinkedListNode<MRUItem> node = FindNode(value);
            if (node != null)
            {
                items.Remove(node);
                items.AddFirst(node);
                OnOrderChanged();
                ++version;
                return node.Value;
            }
            return null;
        }

        private LinkedListNode<MRUItem> FindNode(TValue value)
        {
            LinkedListNode<MRUItem> node = items.First;
            while (node != null)
            {
                if (isEqualFunc(value, node.Value.Value))
                    return node;
                node = node.Next;
            }
            return null;
        }

        #region Registry
        public bool SaveToRegistry(string rootKey)
        {
            int n = 1;
            LinkedListNode<MRUItem> node = items.First;
            Registry.SetValue(rootKey, "Items", MaxSize);
            while (node != null)
            {
                Registry.SetValue(rootKey, String.Format("Item{0}", n), node.Value.Value);
                n++;
                node = node.Next;
            }
            return true;
        }

        static public MRUList<TValue> LoadFromRegistry(string rootKey, MRUEqualityComparer func)
        {
            int n = 1, sz;
            object o;
            o = Registry.GetValue(rootKey, "Items", 8);
            if (o == null)
                sz = 8;
            else
                sz = (int)o;

            MRUList<TValue> newList = new MRUList<TValue>(sz, func);
            while (n <= sz)
            {
                TValue val = (TValue)Registry.GetValue(rootKey, String.Format("Item{0}", n), default(TValue));
                if (val == null)
                    break;
                newList.Add(val);
                
                n++;
            }
            return newList;
        }
        #endregion

        //public TValue this[int index]
        //{
        //    get
        //    {
        //        return items[index];
        //    }
        //    set
        //    {
        //        ++version;
        //        items[index] = value;
        //    }
        //}


        #region IEnumerable<TValue> Members

        public IEnumerator<TValue> GetEnumerator()
        {
            return new Enumerator(this);
        }

        #endregion

        #region IEnumerable Members

        System.Collections.IEnumerator System.Collections.IEnumerable.GetEnumerator()
        {
            return new Enumerator(this);
        }

        #endregion

        private class Enumerator : IEnumerator<TValue>
        {
            private int version;
            private LinkedListNode<MRUItem> cursor;
            private MRUList<TValue> oThis;

            public Enumerator(MRUList<TValue> obj)
            {
                oThis = obj;
                cursor = null;
                version = oThis.version;
            }

            public TValue Current
            {
                get
                {
                    if (oThis.version != version)
                    {
                        throw new InvalidOperationException("Collection was modified");
                    }
                    if (cursor == null)
                    {
                        throw new InvalidOperationException("Enumeration not started");
                    }
                    return cursor.Value.Value;
                }
            }
 
            public void Dispose()
            {
                cursor = null;
            }
 
            public bool MoveNext()
            {
                if (cursor == null)
                    cursor = oThis.items.First;
                else 
                    cursor = cursor.Next;

                if (cursor == null)
                {
                    return false;
                }
                return true;
            }
 
            public void Reset()
            {
                cursor = null;
            }
 
            object IEnumerator.Current
            {
                get
                {
                    return Current;
                }
            }
        }
    }
}
